import sys
from datetime import date, datetime
from decimal import Decimal
# unused scalar elements: INTERVAL, JSON, NUMERIC, PG_JSONB, PG_NUMERIC, PG_OID
from google.cloud.spanner_v1.param_types import (
    Type, BOOL, BYTES, DATE, FLOAT32, FLOAT64, INT64, STRING, TIMESTAMP
)
from pypomes_core import env_get_str, exc_format, str_splice
from typing import Any

from .db_common import (
    _DB_CONN_DATA, _DB_LOGGERS, DbEngine, _assert_query_quota
)
from .db_pomes import db_add_query_limits
from .spanner_frame import SpannerConnection, SpannerParam


def spanner_setup(instance_id: str,
                  database_id: str,
                  session_pool_size: int = 10,
                  session_ping_interval: int = 3000,
                  session_default_timeout: int = 10) -> bool:
    """
    Establish the configuration parameters for access to the Google Cloud Spanner engine.

    If the Cloud Spanner's local emulator is being used, no session usage is possible, and consequently,
    no session pools are available. The standard way to enable local emulation is to set the environment
    variable *SPANNER_EMULATOR_HOST* to its local IP address (e.g., *http://localhost:9010*).

    :param instance_id: the instance identification
    :param database_id: the databae identification
    :param session_pool_size: size of the session pool (defaults to 10 for cloud operation)
    :param session_ping_interval: interval at which to ping sessions in the background (defaults to 3000 seconds)
    :param session_default_timeout: time to wait for a returned session (defaults to 10 seconds)
    :return: *True* if the data was accepted, *False* otherwise
    """
    # initialize the return variable
    result: bool = False

    emulated: bool = bool(env_get_str(key="SPANNER_EMULATOR_HOST"))

    # accept data only if GoogleSpanner has not been instantiated
    if not (_DB_CONN_DATA.get(DbEngine.SPANNER) or {}).get(SpannerParam.ENGINE):
        _DB_CONN_DATA[DbEngine.SPANNER] = {
            SpannerParam.DATABASE_ID: database_id,
            SpannerParam.INSTANCE_ID: instance_id,
            SpannerParam.SESSION_POOL_SIZE: 0 if emulated else session_pool_size,
            SpannerParam.SESSION_PING_INTERVAL: 0 if emulated else session_ping_interval,
            SpannerParam.SESSION_DEFAULT_TIMEOUT: 0 if emulated else session_default_timeout
        }
        result = True

    return result


def connect(autocommit: bool = None,
            errors: list[str] = None) -> SpannerConnection | None:
    """
    Obtain and return a connection to the database.

    Return *None* if the connection could not be obtained.

    :param autocommit: whether the connection is to be in autocommit mode (defaults to *False*)
    :param errors: incidental error messages (must be *[]* or *None*)
    :return: the connection to the database, or *None* if error
    """
    # initialize the return variable
    result: SpannerConnection | None = None

    try:
        result = SpannerConnection(errors=errors,
                                   autocommit=autocommit)
    except Exception as e:
        exc_err: str = exc_format(exc=e,
                                  exc_info=sys.exc_info())
        if _DB_LOGGERS[DbEngine.SPANNER]:
            _DB_LOGGERS[DbEngine.SPANNER].error(msg=exc_err)
        if isinstance(errors, list):
            errors.append(exc_err)

    return result


def select(sel_stmt: str,
           where_vals: tuple | None,
           min_count: int | None,
           max_count: int | None,
           offset_count: int | None,
           limit_count: int | None,
           connection: SpannerConnection | None,
           committable: bool | None,
           errors: list[str] = None) -> list[tuple] | None:
    """
    Query the database and return all tuples that satisfy the *sel_stmt* command.

    The command can optionally contain selection criteria, with respective values given in *where_vals*.
    If not positive integers, *min_count*, *max_count*, *offset_count*, and *limit_count* are ignored.
    If both *min_count* and *max_count* are specified with equal values, then exactly that number of
    tuples must be returned by the query. The parameter *offset_count* is used to offset the retrieval
    of tuples, and *limit_count* establishes a ceiling on the number of tuples returned.
    If the search is empty, an empty list is returned.

    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param sel_stmt: SELECT command for the search
    :param where_vals: the values to be associated with the selection criteria
    :param min_count: optionally defines the minimum number of tuples expected
    :param max_count: optionally defines the maximum number of expected
    :param offset_count: number of tuples to skip
    :param limit_count: limit to the number of tuples returned, to be specified in the query statement itself
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit operation upon errorless completion
    :param errors: incidental error messages (must be *[]* or *None*)
    :return: list of tuples containing the search result, *[]* on empty search, or *None* if error
    """
    # initialize the return variable
    result: list[tuple] | None = None

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # make sure to have a connection
    curr_conn: SpannerConnection = connection or connect(autocommit=True,
                                                         errors=errors)
    # build the parameter maps
    param_values: dict[str, Any] | None = None
    param_types: dict[str, Type] | None = None
    if not errors and where_vals:
        param_values, param_types = __build_param_maps(values=where_vals,
                                                       errors=errors)
    if not errors:
        # establish offset and limit
        if offset_count or limit_count:
            sel_stmt = db_add_query_limits(sel_stmt=sel_stmt,
                                           offset_count=offset_count,
                                           limit_count=limit_count)
        # retrieve the tuples
        rows: list[tuple] = curr_conn.select(sel_stmt=sel_stmt,
                                             param_values=param_values,
                                             param_types=param_types,
                                             committable=committable,
                                             errors=errors)
        if not errors:
            # log the retrieval operation
            count: int = len(rows)
            if _DB_LOGGERS[DbEngine.POSTGRES]:
                from_table: str = str_splice(sel_stmt + " ",
                                             seps=(" FROM ", " "))[1]
                _DB_LOGGERS[DbEngine.POSTGRES].debug(msg=f"Read {count} tuples from "
                                                         f"{DbEngine.SPANNER}.{from_table}, offset "
                                                         f"{offset_count}, connection {id(curr_conn)}")
            if _assert_query_quota(query=sel_stmt,
                                   engine=DbEngine.SPANNER,
                                   where_vals=where_vals,
                                   count=count,
                                   min_count=min_count,
                                   max_count=max_count,
                                   errors=errors):
                # the query quota was satisfied, return the tuples
                result = rows

    return result


def execute(exc_stmt: str,
            bind_vals: tuple | None,
            return_cols: dict[str, type] | None,
            min_count: int | None,
            max_count: int | None,
            connection: SpannerConnection | None,
            committable: bool | None,
            errors: list[str] = None) -> tuple | int | None:
    """
    Execute the command *exc_stmt* on the database.

    This command might be a DML ccommand modifying the database, such as inserting, updating or
    deleting tuples, or it might be a DDL statement, or it might even be an environment-related command.

    The optional bind values for this operation are in *bind_vals*. The optional *return_cols* indicate that
    the values of the columns therein should be returned upon execution of *exc_stmt*. This is typical for
    *INSERT* or *UPDATE* statements on tables with *identity-type* columns, which are columns whose values
    are generated by the database itself. Otherwise, the value returned is the number of inserted, modified,
    or deleted tuples, or *None* if an error occurred.

    If not positive integers, *min_count* and *max_count* are ignored. If both *min_count* and *max_count*
    are specified with equal values, then exactly that number of tuples must have been affected by the query.

    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param exc_stmt: the command to execute
    :param bind_vals: optional bind values
    :param return_cols: optional columns and respective types, whose values are to be returned on *INSERT* or *UPDATE*
    :param min_count: optionally defines the minimum number of tuples to be affected
    :param max_count: optionally defines the maximum number of tuples to be affected
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (must be *[]* or *None*)
    :return: the values of *return_cols*, the value returned by the operation, or *None* if error
    """
    # initialize the return variable
    result: tuple | int | None = None

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # make sure to have a connection
    curr_conn: SpannerConnection = connection or connect(autocommit=True,
                                                         errors=errors)
    # build the parameter maps
    param_values: dict[str, Any] | None = None
    param_types: dict[str, Type] | None = None
    if not errors and bind_vals:
        param_values, param_types = __build_param_maps(values=bind_vals,
                                                       errors=errors)
    if not errors:
        if return_cols:
            exc_stmt += F" THEN RETURN {', '.join(return_cols.keys())}"

        # execte the opration
        reply: tuple | int = curr_conn.execute(exc_stmt=exc_stmt,
                                               param_values=param_values,
                                               param_types=param_types,
                                               committable=committable,
                                               errors=errors)
        if not errors:
            if return_cols:
                result = reply
            elif isinstance(reply, int) and _assert_query_quota(query=exc_stmt,
                                                                engine=DbEngine.SPANNER,
                                                                where_vals=bind_vals,
                                                                count=reply,
                                                                min_count=min_count,
                                                                max_count=max_count,
                                                                errors=errors):
                # the query quota was satisfied
                result = reply

    return result


def __build_param_maps(values: tuple,
                       errors: list[str]) -> tuple[dict[str, Any], dict[str, Type]]:
    """
    Build the mappings of parameters to values and types, as per the Cloud Spanner standard.

    :param values: values to be used
    :param errors: incidental error messages
    :return: mappings of params to values and params to types
    """
    # initialize the return variables
    param_values: dict[str, Any] = {}
    param_types: dict[str, Type] = {}

    for idx, val in enumerate(iterable=values,
                              start=1):
        key: str = f"gs{idx}"
        param_values[key] = val
        if isinstance(val, str):
            param_types[key] = STRING
        elif isinstance(val, bytes):
            param_types[key] = BYTES
        elif isinstance(val, float):
            param_types[key] = FLOAT32
        elif isinstance(val, Decimal):
            param_types[key] = FLOAT64

        # HAZARD: must test for 'bool' first ('bool' is subclass of 'int')
        elif isinstance(val, bool):
            param_types[key] = BOOL
        elif isinstance(val, int):
            param_types[key] = INT64

        # HAZARD: must test for 'datetime' first ('datetime' is subclass of 'date')
        elif isinstance(val, datetime):
            param_types[key] = TIMESTAMP
        elif isinstance(val, date):
            param_types[key] = DATE

        else:
            msg: str = f"Unable to handle type '{val.__class__}"
            if _DB_LOGGERS[DbEngine.SPANNER]:
                _DB_LOGGERS[DbEngine.SPANNER].error(msg=msg)
            errors.append(msg)

    return param_values, param_types



import pyodbc
from contextlib import suppress
from datetime import date, datetime
from pathlib import Path
from pyodbc import Connection, Row
from pypomes_core import (
    DateFormat, DatetimeFormat,  str_between, str_splice
)
from typing import Any, BinaryIO, Final

from .db_common import (
    _DB_LOGGERS, DbEngine, DbParam,
    _assert_query_quota, _build_query_msg, _get_params, _except_msg
)

RESERVED_WORDS: Final[list[str]] = [
    "ADD", "ALL", "ALTER", "AND", "ANY", "AS", "ASC", "AUTHORIZATION", "BACKUP", "BEGIN", "BETWEEN",
    "BREAK", "BROWSE", "BULK", "BY", "CASE", "CHECK", "CHECKPOINT", "CLOSE", "CLUSTERED", "COALESCE",
    "COLUMN", "COMMIT", "COMPUTE", "CONSTRAINT", "CONTAINS", "CONTAINSTABLE", "CONTINUE", "CONVERT",
    "CREATE", "CROSS", "CURRENT", "CURRENT_DATE", "CURRENT_TIME", "CURRENT_TIMESTAMP", "CURRENT_USER",
    "CURSOR", "DATABASE", "DBCC", "DEALLOCATE", "DECLARE", "DEFAULT", "DELETE", "DENY", "DESC", "DISK",
    "DISTINCT", "DISTRIBUTED", "DOUBLE", "DROP", "DUMP", "ELSE", "END", "ERRLVL", "ESCAPE", "EXCEPT",
    "EXEC", "EXECUTE", "EXISTS", "EXIT", "EXTERNAL", "FETCH", "FILE", "FILLFACTOR", "FOR", "FOREIGN",
    "FREETEXT", "FREETEXTTABLE", "FROM", "FULL", "FUNCTION", "GOTO", "GRANT", "GROUP", "HAVING", "HOLDLOCK",
    "IDENTITY", "IDENTITY_INSERT", "IDENTITYCOL", "IF", "ILIKE", "IN", "INDEX", "INNER", "INSERT", "INTERSECT",
    "INTO", "IS", "JOIN", "KEY", "KILL", "LEFT", "LIKE", "LINENO", "LOAD", "MERGE", "NATIONAL", "NOCHECK",
    "NONCLUSTERED", "NOT", "NULL", "NULLIF", "OF", "OFF", "OFFSETS", "ON", "OPEN", "OPENDATASOURCE",
    "OPENQUERY", "OPENROWSET", "OPENXML", "OPTION", "OR", "ORDER", "OUTER", "OVER", "PERCENT", "PIVOT",
    "PLAN", "PRECISION", "PRIMARY", "PRINT", "PROC", "PROCEDURE", "PUBLIC", "RAISERROR", "READ",
    "READTEXT", "RECONFIGURE", "REFERENCES", "REPLICATION", "RESTORE", "RESTRICT", "RETURN", "REVERT",
    "REVOKE", "RIGHT", "ROLLBACK", "ROWCOUNT", "ROWGUIDCOL", "RULE", "SAVE", "SCHEMA", "SECURITYAUDIT",
    "SELECT", "SEMANTICKEYPHRASETABLE", "SEMANTICSIMILARITYDETAILSTABLE", "SEMANTICSIMILARITYTABLE",
    "SESSION_USER", "SET", "SETUSER", "SHUTDOWN", "SOME", "STATISTICS", "SYSTEM_USER", "TABLE",
    "TABLESAMPLE", "TEXTSIZE", "THEN", "TO", "TOP", "TRAN", "TRANSACTION", "TRIGGER", "TRUNCATE",
    "TRY_CONVERT", "TSEQUAL", "UNION", "UNIQUE", "UNPIVOT", "UPDATE", "UPDATETEXT", "USE", "USER",
    "VALUES", "VARYING", "VIEW", "WAITFOR", "WHEN", "WHERE", "WHILE", "WITH", "WITHIN GROUP", "WRITETEXT"
]


def get_connection_string() -> str:
    """
    Build and return the connection string for connecting to the database.

    :return: the connection string
    """
    # retrieve the connection parameters
    db_params: dict[DbParam, Any] = _get_params(DbEngine.SQLSERVER)

    # build and return the connection string
    return (
        f"mssql+pyodbc://{db_params.get(DbParam.USER)}:"
        f"{db_params.get(DbParam.PWD)}@{db_params.get(DbParam.HOST)}:"
        f"{db_params.get(DbParam.PORT)}/{db_params.get(DbParam.NAME)}?driver={db_params.get(DbParam.DRIVER)}"
    )


def connect(autocommit: bool = None,
            errors: list[str] = None) -> Connection | None:
    """
    Obtain and return a connection to the database.

    Return *None* if the connection could not be obtained.

    :param autocommit: whether the connection is to be in autocommit mode (defaults to *False*)
    :param errors: incidental error messages (must be *[]* or *None*)
    :return: the connection to the database, or *None* if error
    """
    # initialize the return valiable
    result: Connection | None = None

    # retrieve the connection parameters and build the connection string
    db_params: dict[DbParam, Any] = _get_params(DbEngine.SQLSERVER)
    connection_kwargs: str = (
        f"DRIVER={{{db_params.get(DbParam.DRIVER)}}};"
        f"SERVER={db_params.get(DbParam.HOST)},{db_params.get(DbParam.PORT)};"
        f"DATABASE={db_params.get(DbParam.NAME)};"
        f"UID={db_params.get(DbParam.USER)};PWD={db_params.get(DbParam.PWD)};TrustServerCertificate=yes;"
    )

    # obtain a connection to the database
    try:
        result = pyodbc.connect(connection_kwargs)
        # establish the connection's autocommit mode
        result.autocommit = isinstance(autocommit, bool) and autocommit
    except Exception as e:
        msg: str = _except_msg(exception=e,
                               connection=None,
                               engine=DbEngine.SQLSERVER)
        if _DB_LOGGERS[DbEngine.SQLSERVER]:
            _DB_LOGGERS[DbEngine.SQLSERVER].error(msg=msg)
        if isinstance(errors, list):
            errors.append(msg)

    # log errors
    if _DB_LOGGERS[DbEngine.SQLSERVER] and errors:
        _DB_LOGGERS[DbEngine.SQLSERVER].error(msg=f"Error connecting to '{db_params.get(DbParam.NAME)}' "
                                                  f"at '{db_params.get(DbParam.HOST)}'")
    return result


def bind_arguments(stmt: str,
                   bind_vals: list[Any]) -> str:
    """
    Replace the placeholders in *query_stmt* with the values in *bind_vals*, and return the modified query statement.

    Note that using a statement in a situation where values for types other than *bool*, *str*, *int*, *float*,
    *date*, or *datetime* were replaced, may bring about undesirable consequences, as the standard string
    representations for these other types would be used.

    The third parameter in the *CONVERT()* function is the style code to be used. Refer to SQLServer's
    documentation for details about this function.

    :param stmt: the query statement
    :param bind_vals: the values to replace the placeholders with
    :return: the query statement with the placeholders replaced with their corresponding values
    """
    # initialize the return variable
    result: str = stmt

    # bind the arguments
    for bind_val in bind_vals:
        val: str
        if isinstance(bind_val, bool):
            val = "1" if bind_val else "0"
        elif isinstance(bind_val, int | float):
            val = f"{bind_val}"
        elif isinstance(bind_val, date):
            val = f"CONVERT(DATE, '{bind_val.strftime(format=DateFormat.INV)}', 23)"
        elif isinstance(bind_val, datetime):
            val = f"CONVERT(DATETIME, '{bind_val.strftime(format=DatetimeFormat.INV)}', 120)"
        else:
            val = f"'{bind_val}'"
        result = result.replace("?", val, 1)

    return result


def select(sel_stmt: str,
           where_vals: tuple | None,
           min_count: int | None,
           max_count: int | None,
           offset_count: int | None,
           limit_count: int | None,
           connection: Connection | None,
           committable: bool | None,
           errors: list[str] = None) -> list[tuple] | None:
    """
    Query the database and return all tuples that satisfy the *sel_stmt* command.

    The command can optionally contain selection criteria, with respective values given in *where_vals*.
    If not positive integers, *min_count*, *max_count*, *offset_count*, and *limit_count* are ignored.
    If both *min_count* and *max_count* are specified with equal values, then exactly that number of
    tuples must be returned by the query. The parameter *offset_count* is used to offset the retrieval
    of tuples. For both *offset_count* and *limit_count* to be used together, an *ORDER BY* clause must
    have been specifed, otherwise a runtime error is raised. If the search is empty, an empty list is returned.

    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param sel_stmt: SELECT command for the search
    :param where_vals: the values to be associated with the selection criteria
    :param min_count: optionally defines the minimum number of tuples expected
    :param max_count: optionally defines the maximum number of tuples expected
    :param offset_count: number of tuples to skip
    :param limit_count: limit to the number of tuples returned, to be specified in the query statement itself
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (must be *[]* or *None*)
    :return: list of tuples containing the search result, *[]* on empty search, or *None* if error
    """
    # initialize the return variable
    result: list[tuple] | None = None

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # make sure to have a connection
    curr_conn: Connection = connection or connect(autocommit=False,
                                                  errors=errors)
    if not errors:
        # establish offset and limit
        if offset_count or limit_count:
            sel_stmt = add_query_limits(sel_stmt=sel_stmt,
                                        offset_count=offset_count,
                                        limit_count=limit_count)
        try:
            # obtain a cursor and execute the operation
            with curr_conn.cursor() as cursor:
                if where_vals:
                    cursor.execute(sel_stmt,
                                   where_vals)
                else:
                    cursor.execute(sel_stmt)
                rows: list[Row] = cursor.fetchall()
                # obtain the number of tuples returned
                count: int = len(rows)

                # log the retrieval operation
                if _DB_LOGGERS[DbEngine.SQLSERVER]:
                    from_table: str = str_splice(sel_stmt + " ",
                                                 seps=(" FROM ", " "))[1]
                    _DB_LOGGERS[DbEngine.SQLSERVER].debug(msg=f"Read {count} tuples from "
                                                              f"{DbEngine.SQLSERVER}.{from_table}, offset "
                                                              f"{offset_count}, connection {id(curr_conn)}")

                if _assert_query_quota(query=sel_stmt,
                                       engine=DbEngine.SQLSERVER,
                                       where_vals=where_vals,
                                       count=count,
                                       min_count=min_count,
                                       max_count=max_count,
                                       errors=errors):
                    # the query quota been satisfied, retrieve the returned tuples
                    result = [tuple(row) for row in rows]

            # commit the transaction
            if committable or not connection:
                curr_conn.commit()
        except Exception as e:
            if committable or not connection:
                with suppress(Exception):
                    curr_conn.rollback()
            msg: str = _except_msg(exception=e,
                                   connection=curr_conn,
                                   engine=DbEngine.SQLSERVER)
            if _DB_LOGGERS[DbEngine.SQLSERVER]:
                _DB_LOGGERS[DbEngine.SQLSERVER].error(msg=msg)
            if isinstance(errors, list):
                errors.append(msg)
        finally:
            # close the connection, if locally acquired
            if curr_conn and not connection:
                curr_conn.close()
        # log errors
        if _DB_LOGGERS[DbEngine.SQLSERVER] and errors:
            _DB_LOGGERS[DbEngine.SQLSERVER].error(msg=_build_query_msg(query_stmt=sel_stmt,
                                                  engine=DbEngine.SQLSERVER,
                                                  bind_vals=where_vals))
    return result


def execute(exc_stmt: str,
            bind_vals: tuple | None,
            return_cols: dict[str, type] | None,
            min_count: int | None,
            max_count: int | None,
            connection: Connection | None,
            committable: bool | None,
            errors: list[str] = None) -> tuple | int | None:
    """
    Execute the command *exc_stmt* on the database.

    This command might be a DML ccommand modifying the database, such as inserting, updating or
    deleting tuples, or it might be a DDL statement, or it might even be an environment-related command.

    The optional bind values for this operation are in *bind_vals*. The optional *return_cols* indicate that
    the values of the columns therein should be returned upon execution of *exc_stmt*. This is typical for
    *INSERT* or *UPDATE* statements on tables with *identity-type* columns, which are columns whose values
    are generated by the database itself. Otherwise, the value returned is the number of inserted, modified,
    or deleted tuples, or *None* if an error occurred.

    If not positive integers, *min_count* and *max_count* are ignored. If both *min_count* and *max_count*
    are specified with equal values, then exactly that number of tuples must have been affected by the query,
    as reported by *cursor.rowcount*.

    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param exc_stmt: the command to execute
    :param bind_vals: optional bind values
    :param return_cols: optional columns and respective types, whose values are to be returned on *INSERT* or *UPDATE*
    :param min_count: optionally defines the minimum number of tuples to be affected
    :param max_count: optionally defines the maximum number of tuples to be affected
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (must be *[]* or *None*)
    :return: the values of *return_cols*, the value returned by the operation, or *None* if error
    """
    # initialize the return variable
    result: tuple | int | None = None

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # make sure to have a connection
    curr_conn: Connection = connection or connect(autocommit=False,
                                                  errors=errors)
    if not errors:
        # handle return columns
        if return_cols:
            # 'DELETED' and 'INSERTED' refer to the old and new values, respectively
            cols: list[str] = [f"INSERTED.{col}" for col in return_cols]
            pos: int = exc_stmt.find(" VALUES(")
            if pos < 0:
                pos = exc_stmt.find(" WHERE ")
            if pos > 0:
                exc_stmt = exc_stmt[:pos] + f" OUTPUT {', '.join(cols)}" + exc_stmt[pos:]
            else:
                exc_stmt += f" OUTPUT {', '.join(cols)}"
        try:
            # obtain a cursor and execute the operation
            with curr_conn.cursor() as cursor:
                # SQLServer understands 'None' value as an effective bind value
                if bind_vals:
                    cursor.execute(exc_stmt,
                                   bind_vals)
                else:
                    cursor.execute(exc_stmt)

                # check whether the query quota has been satisfied
                count: int = cursor.rowcount
                if _assert_query_quota(query=exc_stmt,
                                       engine=DbEngine.SQLSERVER,
                                       where_vals=None,
                                       count=count,
                                       min_count=min_count,
                                       max_count=max_count,
                                       errors=errors):
                    if return_cols:
                        result = tuple(cursor.fetchone())
                    else:
                        result = count

            # commit the transaction
            if committable or not connection:
                curr_conn.commit()
        except Exception as e:
            if committable or not connection:
                with suppress(Exception):
                    curr_conn.rollback()
            msg: str = _except_msg(exception=e,
                                   connection=curr_conn,
                                   engine=DbEngine.SQLSERVER)
            if _DB_LOGGERS[DbEngine.SQLSERVER]:
                _DB_LOGGERS[DbEngine.SQLSERVER].error(msg=msg)
            if isinstance(errors, list):
                errors.append(msg)
        finally:
            # close the connection, if locally acquired
            if curr_conn and not connection:
                curr_conn.close()
        # log errors
        if _DB_LOGGERS[DbEngine.SQLSERVER] and errors:
            _DB_LOGGERS[DbEngine.SQLSERVER].error(msg=_build_query_msg(query_stmt=exc_stmt,
                                                                       engine=DbEngine.SQLSERVER,
                                                                       bind_vals=bind_vals))
    return result


def bulk_execute(exc_stmt: str,
                 exc_vals: list[tuple],
                 connection: Connection | None,
                 committable: bool | None,
                 errors: list[str] = None) -> int | None:
    """
    Bulk-update the database with the statement defined in *execute_stmt*, and the values in *execute_vals*.

    The binding is done by position. Thus, the binding clauses in *execute_stmt* must contain as many '?'
    placeholders as there are elements in the tuples found in the list provided in *execute_vals*.
    Note that, in *UPDATE* operations, the placeholders in the *WHERE* clause will follow
    the ones in the *SET* clause.

    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param exc_stmt: the command to update the database with
    :param exc_vals: the list of values for tuple identification, and to update the database with
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (must be *[]* or *None*)
    :return: the number of inserted or updated tuples, or *None* if error
    """
    # initialize the return variable
    result: int | None = None

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # make sure to have a connection
    curr_conn: Connection = connection or connect(autocommit=False,
                                                  errors=errors)
    if not errors:
        try:
            # obtain a cursor and perform the operation
            with curr_conn.cursor() as cursor:
                cursor.fast_executemany = True
                cursor.executemany(exc_stmt, exc_vals)
                result = len(exc_vals)

            # commit the transaction
            if committable or not connection:
                curr_conn.commit()
        except Exception as e:
            if committable or not connection:
                with suppress(Exception):
                    curr_conn.rollback()
            msg: str = _except_msg(exception=e,
                                   connection=curr_conn,
                                   engine=DbEngine.SQLSERVER)
            if _DB_LOGGERS[DbEngine.SQLSERVER]:
                _DB_LOGGERS[DbEngine.SQLSERVER].error(msg=msg)
            if isinstance(errors, list):
                errors.append(msg)
        finally:
            # close the connection, if locally acquired
            if curr_conn and not connection:
                curr_conn.close()
        if _DB_LOGGERS[DbEngine.SQLSERVER] and errors:
            _DB_LOGGERS[DbEngine.SQLSERVER].error(msg=_build_query_msg(query_stmt=exc_stmt,
                                                                       engine=DbEngine.SQLSERVER,
                                                                       bind_vals=exc_vals[0]))
    return result


def update_lob(lob_table: str,
               lob_column: str,
               pk_columns: list[str],
               pk_vals: tuple,
               lob_data: bytes | str | Path | BinaryIO,
               chunk_size: int,
               connection: Connection | None,
               committable: bool | None,
               errors: list[str] = None) -> None:
    """
    Update a large binary objects (LOB) in the given table and column.

    The data for the update may come from *bytes*, from a *Path* or its string representation,
    or from a pointer obtained from *BytesIO* or *Path.open()* in binary mode.

    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param lob_table: the table to be update with the new LOB
    :param lob_column: the column to be updated with the new LOB
    :param pk_columns: columns making up a primary key, or a unique identifier for the tuple
    :param pk_vals: values with which to locate the tuple to be updated
    :param lob_data: the LOB data (bytes, a file path, or a file pointer)
    :param chunk_size: size in bytes of the data chunk to read/write, or 0 or *None* for no limit
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (must be *[]* or *None*)
    """
    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # make sure to have a connection
    curr_conn: Connection = connection or connect(autocommit=False,
                                                  errors=errors)
    if not errors:
        if isinstance(lob_data, str):
            lob_data = Path(lob_data)

        # normalize the chunk size
        if not chunk_size:
            chunk_size = -1

        # build the UPDATE query
        where_clause: str = " AND ".join([f"{column} = ?" for column in pk_columns])
        update_stmt: str = (f"UPDATE {lob_table} "
                            f"SET {lob_column} = ? "
                            f"WHERE {where_clause}")
        try:
            # obtain a cursor and execute the operation
            with curr_conn.cursor() as cursor:

                # retrieve the lob data and write to the database
                # ('pyodbc.Binary' is an alias for 'bytes')
                # from pyodbc import Binary
                if isinstance(lob_data, bytes):
                    # cursor.execute(update_stmt,
                    #                (Binary(lob_data), *pk_vals))
                    cursor.execute(update_stmt,
                                   (lob_data, *pk_vals))
                elif isinstance(lob_data, Path):
                    data_bytes: bytes
                    with lob_data.open("rb") as file:
                        data_bytes = file.read(chunk_size)
                        while data_bytes:
                            # cursor.execute(update_stmt,
                            #                (Binary(data_bytes), *pk_vals))
                            cursor.execute(update_stmt,
                                           (data_bytes, *pk_vals))
                            data_bytes = file.read(chunk_size)
                else:
                    data_bytes: bytes = lob_data.read(chunk_size)
                    while data_bytes:
                        # cursor.execute(update_stmt,
                        #                (Binary(data_bytes), *pk_vals))
                        cursor.execute(update_stmt,
                                       (data_bytes, *pk_vals))
                        data_bytes = lob_data.read(chunk_size)
                    lob_data.close()

            # commit the transaction
            if committable or not connection:
                curr_conn.commit()
        except Exception as e:
            if committable or not connection:
                with suppress(Exception):
                    curr_conn.rollback()
            msg: str = _except_msg(exception=e,
                                   connection=curr_conn,
                                   engine=DbEngine.SQLSERVER)
            if _DB_LOGGERS[DbEngine.SQLSERVER]:
                _DB_LOGGERS[DbEngine.SQLSERVER].error(msg=msg)
            if isinstance(errors, list):
                errors.append(msg)
        finally:
            # close the connection, if locally acquired
            if curr_conn and not connection:
                curr_conn.close()
        # log errors
        if _DB_LOGGERS[DbEngine.SQLSERVER] and errors:
            _DB_LOGGERS[DbEngine.SQLSERVER].error(msg=_build_query_msg(query_stmt=update_stmt,
                                                                       engine=DbEngine.SQLSERVER,
                                                                       bind_vals=pk_vals))


def add_query_limits(sel_stmt: str,
                     offset_count: int | None,
                     limit_count: int | None) -> str:
    """
    Add offset and limit to *sel_stmt*.

    If not positive integers, *offset_count*, and *limit_count* are ignored. The parameter *offset_count*
    is used to offset the retrieval of tuples, whereas *limit_count* establishes a ceiling on the number
    of tuples returned.

    :param sel_stmt: the query statement to modify
    :param offset_count: number of tuples to skip
    :param limit_count: limit to the number of tuples returned
    :return: the modified query statement
    """
    # initialize the return variable
    result: str = sel_stmt

    # establish offset and limit for query (TOP and OFFSET clauses cannot be used together)
    if isinstance(offset_count, int) and offset_count > 0:
        # establish an offset into the result set
        result += f" OFFSET {offset_count} ROWS"
        if isinstance(limit_count, int) and limit_count > 0:
            # establish a limit to the number of tuples returned
            result += f" FETCH NEXT {limit_count} ROWS ONLY"
    elif isinstance(limit_count, int) and limit_count > 0:
        # establish a limit to the number of tuples returned
        if "SELECT DISTINCT " in result:
            # 1. to get the top N distinct rows:
            #    'DISTINCT' is applied before the 'TOP' clause -
            #    the top N rows are selected, and then duplicates are removed
            # 2. to get the distinct rows first, and then pick the top N from that set:
            #    a subquery has to be used -
            #    SELECT TOP (N) *
            #    FROM (
            #        SELECT DISTINCT <coumns>
            #        FROM <table
            #    ) AS distinct_rows
            #    ...
            result = result.replace("SELECT DISTINCT ", f"SELECT DISTINCT TOP {limit_count} ", 1)
        else:
            result = result.replace("SELECT ", f"SELECT TOP {limit_count} ", 1)

    return result


def call_procedure(proc_name: str,
                   proc_vals: tuple | None,
                   connection: Connection | None,
                   committable: bool | None,
                   errors: list[str] = None) -> list[tuple] | None:
    """
    Execute the stored procedure *proc_name* in the database, with the parameters given in *proc_vals*.

    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param proc_name: name of the stored procedure
    :param proc_vals: parameters for the stored procedure
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (must be *[]* or *None*)
    :return: the data returned by the procedure, or *None* if error
    """
    # initialize the return variable
    result: list[tuple] | None = None

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # make sure to have a connection
    curr_conn: Connection = connection or connect(autocommit=False,
                                                  errors=errors)
    if not errors:
        # build the command
        proc_stmt: str | None = None

        # execute the stored procedure
        try:
            # obtain a cursor and execute the operation
            with curr_conn.cursor() as cursor:
                proc_stmt = f"SET NOCOUNT ON; EXEC {proc_name} {','.join(('?',) * len(proc_vals))}"
                cursor.execute(proc_stmt,
                               proc_vals)
                # retrieve the returned tuples
                rows: list[Row] = cursor.fetchall()
                result = [tuple(row) for row in rows]

            # commit the transaction
            if committable or not connection:
                curr_conn.commit()
        except Exception as e:
            if committable or not connection:
                with suppress(Exception):
                    curr_conn.rollback()
            msg: str = _except_msg(exception=e,
                                   connection=curr_conn,
                                   engine=DbEngine.SQLSERVER)
            if _DB_LOGGERS[DbEngine.SQLSERVER]:
                _DB_LOGGERS[DbEngine.SQLSERVER].error(msg=msg)
            if isinstance(errors, list):
                errors.append(msg)
        finally:
            # close the connection, if locally acquired
            if curr_conn and not connection:
                curr_conn.close()
        # log errors
        if _DB_LOGGERS[DbEngine.SQLSERVER] and errors:
            _DB_LOGGERS[DbEngine.SQLSERVER].error(msg=_build_query_msg(query_stmt=proc_stmt,
                                                                       engine=DbEngine.SQLSERVER,
                                                                       bind_vals=proc_vals))
    return result


def identity_pre_insert(insert_stmt: str,
                        connection: Connection,
                        errors: list[str] = None) -> None:
    """
    Handle the pre-insert for tables with identity columns.

    Identity columns are columns whose values are generated directly by the database engine,
    and as such, require special handling before and after bulk inserts.

    :param insert_stmt: the INSERT command
    :param connection: the connection to use
    :param errors: incidental error messages (must be *[]* or *None*)
    """
    table_name: str = str_between(insert_stmt.upper(),
                                  from_str=" INTO ",
                                  to_str=" ")
    execute(exc_stmt=f"SET IDENTITY_INSERT {table_name.lower()} ON",
            bind_vals=None,
            return_cols=None,
            min_count=None,
            max_count=None,
            connection=connection,
            committable=False,
            errors=errors)


def identity_post_insert(insert_stmt: str,
                         connection: Connection,
                         committable: bool,
                         identity_column: str,
                         errors: list[str] = None) -> None:
    """
    Handle the post-insert for tables with identity columns.

    Identity columns are columns whose values are generated directly by the database engine,
    and as such, require special handling before and after bulk inserts.

    :param insert_stmt: the INSERT command
    :param connection: the connection to use
    :param committable: whether to commit or rollback the operation, upon completion
    :param identity_column: column whose values are generated by the database
    :param errors: incidental error messages (must be *[]* or *None*)
    """
    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # obtain the maximum value inserted
    table_name: str = str_between(insert_stmt.upper(),
                                  from_str=" INTO ",
                                  to_str=" ")
    recs: list[tuple[int]] = select(sel_stmt=(f"SELECT MAX({identity_column}) "
                                              f"FROM {table_name}"),
                                    where_vals=None,
                                    min_count=None,
                                    max_count=None,
                                    offset_count=None,
                                    limit_count=None,
                                    connection=connection,
                                    committable=False,
                                    errors=errors)
    if not errors:
        execute(exc_stmt=f"SET IDENTITY_INSERT {table_name} OFF",
                bind_vals=None,
                return_cols=None,
                min_count=None,
                max_count=None,
                connection=connection,
                committable=False,
                errors=errors)
        if not errors:
            execute(exc_stmt=f"DBCC CHECKIDENT ('{table_name}', RESEED, {recs[0][0]})",
                    bind_vals=None,
                    return_cols=None,
                    min_count=None,
                    max_count=None,
                    connection=connection,
                    committable=committable,
                    errors=errors)



from contextlib import suppress
from logging import Logger
from typing import Any

from .db_common import (
    _DB_LOGGERS, DB_BIND_META_TAG, DbEngine,
    _assert_engine, _except_msg, _remove_ctrlchars
)
from .db_pomes import (
    db_connect, db_select, db_count, db_close,
    db_bulk_delete, db_bulk_insert, db_bulk_update
)


# ruff: noqa: PLR0915 - too many statements
def db_sync_data(source_engine: DbEngine,
                 source_table: str,
                 target_engine: DbEngine,
                 target_table: str,
                 pk_columns: list[str],
                 sync_columns: list[str],
                 identity_column: str = None,
                 ignore_updates: bool = None,
                 offset_count: int = None,
                 limit_count: int = None,
                 batch_size: int = None,
                 has_nulls: bool = None,
                 source_conn: Any = None,
                 source_committable: bool = None,
                 target_conn: Any = None,
                 target_committable: bool = None,
                 errors: list[str] = None,
                 log_trigger: int = 10000) -> tuple[int, int, int] | None:
    """
    Synchronize data in *target_table*, as per the contents of *source_table*.

    The origin and destination databases must be in the list of databases configured and
    supported by this package. The specified database columns must not be type *LOB*
    (binary large object), and must be of respectively equivalent types, in both the origin
    and the destination databases. Specifying a list of one or more columns in *pk_columns*
    that uniquely identify the tuples is required.

    If *ignore_updates* is set to *True*, then the synchronization is limited to *INSERT* and *DELETE*
    operations, and content differences in non-PK attributes are ignored.

    Specific handling is required for identity columns (i.e., columns with values generated
    directly by the database - typically, they are also primary keys), and thus they must be
    identified by *identity_column*. For target Oracle databases, *GENERATED ALWAYS*
    identity columns should be ommited, as attempting to insert values therein raises an error,
    whereas *GENERATED BY DEFAULT AS IDENTITY* columns are acceptable.

    The parameter *has_nulls* indicates that one or more columns holding string content in the source
    table have embedded NULL characters. Some databases, particularly Postgres, do not accept NULLs
    embedded in text, so they must be removed.

    The parameters *source_committable* and *target_committable* define whether a commit or rollback
    is performed on the respective connections, if provided.

    :param source_engine: the source database engine
    :param source_table: the possibly schema-qualified table to read the data from
    :param target_engine: the destination database engine
    :param target_table: the possibly schema-qualified table to write the data to
    :param pk_columns: primary key columns, or a unique identifier for a tuple in both tables
    :param sync_columns: the columns to be synchronized (must not include PK columns)
    :param identity_column: column whose values are generated by the database
    :param ignore_updates: whether to ignore tuple updates (defaults to synchronizing with updates)
    :param offset_count: number of tuples to skip in source table (defaults to none)
    :param limit_count: maximum number of tuples to migrate (defaults to no limit)
    :param batch_size: the maximum number of tuples to synchronize in each batch, or 0 or *None* for no limit
    :param has_nulls: indicates that one or more string-type source columns have nulls
    :param source_conn: the connection to the source database (obtains a new one, if not provided)
    :param source_committable: whether to commit or rollback on *source_conn*, upon completion
    :param target_conn: the connection to the destination database (obtains a new one, if not provided)
    :param target_committable: whether to commit or rollback on *target_conn*, upon completion
    :param errors: incidental error messages (might be a non-empty list)
    :param log_trigger: number of LOBs to trigger logging info on migration (defaults to 10000 LOBs)
    :return: the number of tuples effectively deleted, inserted, and updated in the target table, or *None* if error
    """
    # helper function to obtain a primary key tuple from a row
    def get_pk_data(row: tuple, pk_cols: list[str]) -> tuple:
        return tuple(row[pos] for pos in range(len(pk_cols)))

    # helper function to obtain a sync data tuple from a row
    def get_sync_data(row: tuple, pk_cols: list[str], sync_cols: list[str]) -> tuple:
        return tuple(row[pos] for pos in range(len(pk_cols), len(pk_cols) + len(sync_cols)))

    # initialize the return variable
    result: tuple[int, int, int] | None = None

    # necessary, lest the state of 'errors' be tested
    curr_errors: list[str] = []

    # assert the database engines
    source_engine = _assert_engine(engine=source_engine,
                                   errors=curr_errors)
    target_engine = _assert_engine(engine=target_engine,
                                   errors=curr_errors)
    # obtain the logger
    logger: Logger = _DB_LOGGERS[source_engine] or _DB_LOGGERS[target_engine]

    from_table: str = f"{source_engine}.{source_table}"
    to_table: str = f"{target_engine}.{target_table}"

    curr_source_conn: Any = None
    curr_target_conn: Any = None
    if pk_columns:
        # make sure to have connections to the source and destination databases
        curr_source_conn = source_conn or db_connect(engine=source_engine,
                                                     errors=curr_errors)
        curr_target_conn = target_conn or db_connect(engine=target_engine,
                                                     errors=curr_errors)
    else:
        err_msg: str = (f"Cannot synchronize tables {source_engine}.{from_table} "
                        f"and {target_engine}{to_table}: "
                        f"no columns to uniquely identify tuples specified")
        if logger:
            logger.error(msg=err_msg)
        curr_errors.append(err_msg)

    if curr_source_conn and curr_target_conn:
        # normalize these parameters
        offset_count = __normalize_value(value=offset_count)
        limit_count = __normalize_value(value=limit_count)
        batch_size = __normalize_value(value=batch_size)

        # buid the SELECT query
        source_cols: str = ", ".join(pk_columns + sync_columns)
        source_sel_stmt: str = (f"SELECT {source_cols} FROM {source_table} "
                                f"ORDER BY {', '.join(pk_columns)}")
        target_cols: str = ", ".join(pk_columns) if ignore_updates else source_cols
        target_sel_stmt: str = (f"SELECT {target_cols} FROM {target_table} "
                                f"ORDER BY {', '.join(pk_columns)}")
        if batch_size or offset_count:
            source_sel_stmt += " OFFSET @offset ROWS"
            target_sel_stmt += " OFFSET @offset ROWS"
        if batch_size or limit_count:
            if source_engine == DbEngine.POSTGRES:
                source_sel_stmt += " LIMIT @limit"
            elif source_engine in [DbEngine.ORACLE, DbEngine.SQLSERVER]:
                source_sel_stmt += " FETCH NEXT @limit ROWS ONLY"
            if target_engine == DbEngine.POSTGRES:
                target_sel_stmt += " LIMIT @limit"
            elif target_engine in [DbEngine.ORACLE, DbEngine.SQLSERVER]:
                target_sel_stmt += " FETCH NEXT @limit ROWS ONLY"

        rows_for_delete: list[tuple] = []
        rows_for_insert: list[tuple] = []
        # not used if 'ignore_updates' is set
        rows_for_update: list[tuple] = []

        # log the synchronization start
        if logger:
            logger.debug(msg=f"Started synchronizing data, "
                             f"from {source_engine}.{from_table}, connection {id(curr_source_conn)}, "
                             f"to {target_engine}.{to_table}, connection {id(curr_target_conn)}")
        # set migration vars
        sel_stmt: str
        log_count: int = 0
        log_step: int = 0
        err_msg: str | None = None

        # process the offsets
        source_offset: int = 0
        target_offset: int = 0
        if offset_count > 0:
            # obtain the source PK values at the given offset
            pk_cols: str = ", ".join(pk_columns)
            source_vals: list[tuple] = db_select(sel_stmt=f"SELECT {pk_cols} FROM {source_table} ",
                                                 orderby_clause=pk_cols,
                                                 offset_count=offset_count - 1,
                                                 engine=source_engine,
                                                 connection=source_conn,
                                                 committable=source_committable if source_conn else True,
                                                 errors=curr_errors)
            if source_vals:
                source_offset = offset_count
                # obtain the target offset for the source PK values
                where_clause = __build_where_count(engine=target_engine,
                                                   pk_columns=pk_columns)
                target_offset = db_count(table=source_table,
                                         where_clause=where_clause,
                                         where_vals=tuple(source_vals[0]),
                                         engine=target_engine,
                                         connection=target_conn,
                                         committable=target_committable if target_conn else True,
                                         errors=curr_errors)
        if not curr_errors:
            # establish the initial limit
            curr_limit: int = min(batch_size, limit_count) or batch_size or limit_count

            try:
                # obtain the cursors
                source_cursor: Any = curr_source_conn.cursor()
                target_cursor: Any = curr_target_conn.cursor()

                # execute the queries
                if source_offset:
                    sel_stmt = source_sel_stmt.replace("@offset", str(source_offset))
                else:
                    sel_stmt = source_sel_stmt.replace("OFFSET @offset ROWS ", "")
                if curr_limit:
                    sel_stmt = sel_stmt.replace("@limit", str(curr_limit))
                else:
                    sel_stmt = sel_stmt.replace("LIMIT @limit ", "")
                # parameter: 'statement' (Oracle), 'query' (PostgreSQL), 'sql' (SQLServer)
                source_cursor.execute(sel_stmt)
                source_rows: list[tuple] = source_cursor.fetchall()
                source_count: int = len(source_rows)
                source_offset += source_count
                log_count = source_count
                total_source: int = source_count
                # log query result
                if logger:
                    logger.debug(msg=f"Read {source_count} tuples "
                                     f"from {source_engine}.{from_table}, "
                                     f"offset {source_offset}, limit {curr_limit}, "
                                     f"connection {id(curr_source_conn)}")
                if has_nulls and target_engine == DbEngine.POSTGRES:
                    _remove_ctrlchars(rows=source_rows)

                if target_offset:
                    sel_stmt = target_sel_stmt.replace("@offset", str(target_offset))
                else:
                    sel_stmt = target_sel_stmt.replace("OFFSET @offset ROWS ", "")
                if curr_limit:
                    sel_stmt = sel_stmt.replace("@limit", str(curr_limit))
                else:
                    sel_stmt = sel_stmt.replace("LIMIT @limit ", "")
                # parameter: 'statement' (Oracle), 'query' (PostgreSQL), 'sql' (SQLServer)
                target_cursor.execute(sel_stmt)
                target_rows: list[tuple] = target_cursor.fetchall()
                target_count = len(target_rows)
                target_offset += target_count
                total_target: int = target_count
                # log query result
                if logger:
                    logger.debug(msg=f"Read {target_count} tuples "
                                     f"from {target_engine}.{target_table}, "
                                     f"offset {target_offset}, limit {curr_limit}, "
                                     f"connection {id(curr_target_conn)}")

                # traverse the result set
                source_row: tuple | None = None
                target_row: tuple | None = None
                source_pks: tuple | None = None
                target_pks: tuple | None = None
                while source_rows or target_rows:
                    if source_rows and not source_row:
                        source_row = source_rows.pop(0)
                        source_pks = get_pk_data(row=source_row,
                                                 pk_cols=pk_columns)
                    if target_rows and not target_row:
                        target_row = target_rows.pop(0)
                        target_pks = get_pk_data(row=target_row,
                                                 pk_cols=pk_columns)
                    if source_row and \
                       (not target_row or source_pks < target_pks):
                        # flag row for insertion
                        rows_for_insert.append(source_row)
                        source_row = None
                    elif target_row and \
                            (not source_row or target_pks < source_pks):
                        # flag row for deletion
                        rows_for_delete.append(target_pks)
                        target_row = None
                    else:
                        if not ignore_updates:
                            source_data: tuple = get_sync_data(row=source_row,
                                                               pk_cols=pk_columns,
                                                               sync_cols=sync_columns)
                            target_data: tuple = get_sync_data(row=target_row,
                                                               pk_cols=pk_columns,
                                                               sync_cols=sync_columns)
                            if source_data != target_data:
                                # flag row for update
                                rows_for_update.append(source_data + source_pks)
                        source_row = None
                        target_row = None

                    if source_offset and not source_rows:
                        if limit_count and total_source >= limit_count:
                            # maximum number of tuples already retrieved from source table
                            source_offset = 0
                        else:
                            sel_stmt = source_sel_stmt.replace("@offset", str(source_offset))
                            curr_limit = min(0, batch_size, limit_count - total_source) \
                                if limit_count else batch_size
                            if curr_limit:
                                sel_stmt = sel_stmt.replace("@limit", str(curr_limit))
                            else:
                                sel_stmt = sel_stmt.replace("LIMIT @limit ", "")
                            # parameter: 'statement' (Oracle), 'query' (PostgreSQL), 'sql' (SQLServer)
                            source_cursor.execute(sel_stmt)
                            source_rows = source_cursor.fetchall()
                            if source_rows:
                                source_count = len(source_rows)
                                source_offset += source_count
                                log_count += source_count
                                log_step += source_count
                                total_source += source_count
                                # log query result
                                if logger:
                                    logger.debug(msg=f"Read {source_count} tuples "
                                                     f"from {source_engine}.{from_table}, "
                                                     f"offset {source_offset}, limit {curr_limit}, "
                                                     f"connection {id(curr_source_conn)}")
                                if has_nulls and target_engine == DbEngine.POSTGRES:
                                    _remove_ctrlchars(rows=source_rows)
                            else:
                                # no more tuples in source table
                                source_offset = 0

                    if target_offset and not target_rows:
                        if limit_count and total_target >= limit_count:
                            # maximum number of tuples already retrieved from target table
                            target_offset = 0
                        else:
                            sel_stmt = target_sel_stmt.replace("@offset", str(target_offset))
                            curr_limit = min(0, batch_size, limit_count - total_target) \
                                if limit_count else batch_size
                            if curr_limit:
                                sel_stmt = sel_stmt.replace("@limit", str(curr_limit))
                            else:
                                sel_stmt = sel_stmt.replace("LIMIT @limit ", "")
                            # parameter: 'statement' (Oracle), 'query' (PostgreSQL), 'sql' (SQLServer)
                            target_cursor.execute(sel_stmt)
                            target_rows = target_cursor.fetchall()
                            if target_rows:
                                target_count = len(target_rows)
                                target_offset += target_count
                                total_target += target_count
                                # log query result
                                if logger:
                                    logger.debug(msg=f"Read {target_count} tuples "
                                                     f"from {target_engine}.{target_table}, "
                                                     f"offset {target_offset}, limit {curr_limit},"
                                                     f" connection {id(curr_target_conn)}")
                            else:
                                # no more tuples in target table
                                target_offset = 0

                    # log partial result
                    if logger and log_step >= log_trigger:
                        logger.debug(msg=f"Synchronizing {log_step} tuples, "
                                         f"from {source_engine}.{from_table}, connection {id(curr_source_conn)}, "
                                         f"to {target_engine}.{to_table}, connection {id(curr_target_conn)}")
                        log_step = 0
                # close the cursors
                source_cursor.close()
                target_cursor.close()

                # process the rows flagged for delete
                if rows_for_delete:
                    # delete the rows
                    db_bulk_delete(target_table=target_table,
                                   where_attrs=pk_columns,
                                   where_vals=rows_for_delete,
                                   engine=target_engine,
                                   connection=curr_target_conn,
                                   committable=False,
                                   errors=curr_errors)

                # process the rows flagged for insert
                if not curr_errors and rows_for_insert:
                    # insert the rows
                    db_bulk_insert(target_table=target_table,
                                   insert_attrs=pk_columns + sync_columns,
                                   insert_vals=rows_for_insert,
                                   engine=target_engine,
                                   connection=curr_target_conn,
                                   committable=target_committable if target_conn else True,
                                   identity_column=identity_column,
                                   errors=curr_errors)

                # process the rows flagged for update
                if not curr_errors and rows_for_update:
                    # update the rows
                    db_bulk_update(target_table=target_table,
                                   set_attrs=sync_columns,
                                   where_attrs=pk_columns,
                                   update_vals=rows_for_update,
                                   engine=target_engine,
                                   connection=curr_target_conn,
                                   committable=target_committable if target_conn else True,
                                   errors=curr_errors)
                if curr_errors:
                    log_count = 0
                else:
                    if source_committable or not source_conn:
                        curr_source_conn.commit()
                    if target_committable or not target_conn:
                        curr_target_conn.commit()
                    result = (len(rows_for_delete), len(rows_for_insert), len(rows_for_update))

            except Exception as e:
                # rollback the transactions
                if source_committable or not source_conn:
                    with suppress(Exception):
                        curr_source_conn.rollback()
                if target_committable or not target_conn:
                    with suppress(Exception):
                        curr_target_conn.rollback()
                log_count = 0
                err_msg = _except_msg(exception=e,
                                      connection=curr_source_conn,
                                      engine=source_engine)
            finally:
                # close the connections, if locally acquired
                if curr_source_conn and not source_conn:
                    db_close(connection=curr_source_conn,
                             engine=source_engine)
                if curr_target_conn and not target_conn:
                    db_close(connection=curr_target_conn,
                             engine=target_engine)

        # log the synchronization finish
        if err_msg:
            curr_errors.append(err_msg)
            if logger:
                logger.error(msg=err_msg)
        elif not curr_errors and logger:
            logger.debug(msg=f"Synchronized {log_count} tuples from "
                             f"{source_engine}.{from_table} to {target_engine}.{to_table}, "
                             f"{result[0]} deleted, {result[1]} inserted, {result[2]} updated")

    if curr_errors and isinstance(errors, list):
        errors.extend(curr_errors)

    return result


def __normalize_value(value: int) -> int:
    """
    Normalize *value* to a consistent value.

    :param value: the value to normalized
    :return: the normalized value
    """
    if isinstance(value, int) and \
            not isinstance(value, bool) and value > 0:
        result: int = value
    else:
        result: int = 0

    return result


def __build_where_count(engine: DbEngine,
                        pk_columns: list[str]) -> str:
    """
    Build the *WHERE* clause for the SQL count statement.

    Suppose *pk_columns* is the list *['pk1', 'pk_2'. 'pk_3']*, and the bind placeholder symbol is *%?*.
    The value returned is:
    ::
      - for SQLServer (no support for row comparisons):
        pk_1 < %?
        OR (pk_1 = %? AND pk_2 < %?)
        OR (pk_1 = %? AND pk_2 = %? AND pk_3 < %?)
      - for the other database engines (full support for row comparisons):
        (pk_1, pk_2, pk_3) < (%?, %?, %?)

    Note that, in due time, the placeholder *%?* will be replaced by the the one specific to *engine*
    (namely, *:n*, *?s*, and *?*, respectively for Oracle, PostgreSQL, and MySQL or SQLServer).

    :param engine: the database engine
    :param pk_columns: primery key columns
    :return: the *WHERE* clause for the SQL count statement
    """
    # declare the return variable
    result: str

    if engine == DbEngine.SQLSERVER:
        # SQLServer does not support row comparisons
        result = f"{pk_columns[0]} < {DB_BIND_META_TAG}"
        if len(pk_columns) > 1:
            for i in range(len(pk_columns) - 1):
                pk_prevs: list[str] = [f"{c} = {DB_BIND_META_TAG}" for c in pk_columns[:i+1]]
                result += f" OR ({' AND '.join(pk_prevs)} AND {pk_columns[i+1]} < {DB_BIND_META_TAG})"
    else:
        result = f"({', '.join(pk_columns)}) < ({', '.join([DB_BIND_META_TAG] * len(pk_columns))})"

    return result



import json
import sys
import time
from contextlib import suppress
from enum import StrEnum, auto
from google.cloud import spanner
from google.cloud.spanner import Client
from google.cloud.spanner import AbstractSessionPool, PingingPool
from google.cloud.spanner_v1.database import Database, StreamedResultSet
from google.cloud.spanner_v1.instance import Instance
from google.cloud.spanner_v1.param_types import Type
from google.cloud.spanner_v1.session import Session
from google.cloud.spanner_v1.transaction import Transaction
from google.oauth2.service_account import Credentials
from io import BytesIO
from pathlib import Path
from pypomes_core import file_get_data, exc_format
from threading import Thread
from typing import Any

from .db_common import (
    _DB_CONN_DATA, _DB_LOGGERS, DbEngine, DbParam
)


class SpannerParam(StrEnum):
    """
    Parameters for *GoogleSpanner*, an encapsulation of the Google Cloud Spanner access lifecycle.
    """
    ENGINE: auto()
    CLIENT_ID: auto()
    DATABASE_ID: auto()
    INSTANCE_ID: auto()
    PROJECT_ID: auto()
    SESSION_DEFAULT_TIMEOUT: auto()
    SESSION_PING_INTERVAL: auto()
    SESSION_POOL_SIZE = auto()


class GoogleSpanner:
    """
    An encapsulation of the Google Cloud Spanner access lifecycle.

    *Spanner* uses *Identity and Access Management* (IAM), rather than simple username/password strings, and thus
    the encapsulation replaces traditional RDBMS URL/User patterns with *GCP Resource Hierarchies*, handling
    the *Google Cloud*'s hierarchical structure, namely, **Project/Instance/Database*. The *google-cloud-spanner*
    library manages a connection pool under the hood, and to encapsulate this logic, this class takes care of the
    *Client* object, while providing access to a *Database* object.

    Extensive documentation on the *google-cloud-spanner* package can be found at
    https://docs.cloud.google.com/python/docs/reference/spanner/latest.

    These are the instance variables:
        - *_client*: *google.cloud.spanner.Client* - the Cloud Spanner Client
        - *_credentials*: *dict[str, str]* - the access credentials
        - *_database*: *google.cloud.spanner_v1.database.Database* - database to use
        - *_instance*: *google.cloud.spanner_v1.instance.Instance* - instance to use
        - *_pinger*: *threading.Thread* - thread for background pinging
    """

    def __init__(self,
                 *,
                 project_id: str = None,
                 credentials: dict[str, str] | BytesIO | Path | str | bytes = None,
                 errors: list[str] = None) -> None:
        r"""
        Encapsulate the Google Cloud Spanner lifecycle.

        The nature of access *credentials* depends on its data type:
            - type *dict*: *credentials* is a Python dictionary
            - type *BytesIO*: *credentials* is a byte stream
            - type *Path*: *credentials* is a path to a file holding the data
            - type *bytes*: *credentials* holds the data (used as is)
            - type *str*: *credentials* holds the data (used as utf8-encoded)

        The *google.oauth2.service_account* package requires the credentials sampled below for authentication:
            - "type": "service_account",
            - "project_id": "my-gcp-project-id",
            - "private_key_id": "a1b2c3d4e5f6g7h8i9j0k1l2m3n4o5p6q7r8s9t0",
            - "private_key": "-----BEGIN PRIVATE KEY-----\\nMIIEvgIBADANBhEFAASC+...\\n-----END PRIVATE KEY-----\\n",
            - "client_email": "my-service-account-name@my-gcp-project-id.iam.gserviceaccount.com",
            - "client_id": "12345678901234567890",
            - "auth_uri": "accounts.google.com",
            - "token_uri": "oauth2.googleapis.com",
            - "auth_provider_x509_cert_url": "www.googleapis.com",
            - "client_x509_cert_url": "www.googleapis.com"

        A credentials file for Google Cloud Spanner is a standard *JSON* service account key file containing
        the data described above. The parameter *project_id* is required if *credentials* is not provided,
        otherwise it is ignored.

        :param errors: incidental error messages (might be a non-empty list)
        """
        # instance objects
        self._client: Client | None = None
        self._instance: Instance | None = None
        self._database: Database | None = None
        self._credentials: dict[str, str] | None = None  # access credentials
        self._pinger: Thread | None = None               # thread for background pinging

        err_msg: str | None = None
        client_id: str | None = None
        if DbEngine.SPANNER not in _DB_CONN_DATA:
            err_msg = "Google Cloud Spanner not yet configured"
        elif _DB_CONN_DATA[DbEngine.SPANNER].get(SpannerParam.ENGINE):
            err_msg = "Instance of Google Cloud Spanner already exists"
        elif project_id or credentials:
            # initilize the Cloud Spanner objects

            if credentials:
                if not isinstance(credentials, dict):
                    credentials_bytes: bytes = file_get_data(file_data=credentials)
                    self._credentials = json.loads(s=credentials_bytes)
                client_id = self._credentials.get("client_id")
                project_id = self._credentials["project_id"]
        else:
            err_msg = "Both 'credentials' and 'project_id' must be provided"

        if err_msg:
            if _DB_LOGGERS[DbEngine.SPANNER]:
                _DB_LOGGERS[DbEngine.SPANNER].error(msg=err_msg)
            if isinstance(errors, list):
                errors.append(err_msg)
        else:
            _DB_CONN_DATA[DbEngine.SPANNER].update({
                SpannerParam.ENGINE: self,
                SpannerParam.CLIENT_ID: client_id,
                SpannerParam.PROJECT_ID: project_id
            })

    def initialize(self,
                   errors: list[str] = None) -> bool:
        """
        Initialize access to the Google Cloud Spanner database engine.

        This method should be invoked once, immediately after object instantiation time.

        :param errors: incidental error messages (might be a non-empty list)
        """
        # initialize the return variable
        result: bool = True

        # obtain the Cloud Spanner configuration data
        conn_data: dict[str, Any] = _DB_CONN_DATA.get(DbEngine.SPANNER)

        err_msg: str | None = None
        if not conn_data or \
           not conn_data.get(SpannerParam.DATABASE_ID) or \
           not conn_data.get(SpannerParam.INSTANCE_ID):
            err_msg = "Unable to obtain session configuratiom parameters"
        elif not conn_data.get(DbParam.VERSION):
            try:
                # obtain a Cloud Spanner Client
                if self._credentials:
                    # use specific service account data
                    credentials: Any = Credentials.from_service_account_info(info=self._credentials)
                    self._client = spanner.Client(project=conn_data[SpannerParam.PROJECT_ID],
                                                  credentials=credentials)
                else:
                    # fallback to Application Default Credentials (ADC)
                    self._client = spanner.Client(project=conn_data[SpannerParam.PROJECT_ID])

                # establish the instance
                self._instance = self._client.instance(instance_id=conn_data[SpannerParam.INSTANCE_ID])

                if conn_data[SpannerParam.SESSION_POOL_SIZE] > 0:
                    # use PingingPool for mixed read/write workloads
                    session_pool: PingingPool = PingingPool(
                        size=conn_data[SpannerParam.SESSION_POOL_SIZE],
                        default_timeout=conn_data[SpannerParam.SESSION_DEFAULT_TIMEOUT],
                        ping_interval=conn_data[SpannerParam.SESSION_PING_INTERVAL],
                    )
                    self._database = self._instance.database(database_id=conn_data[SpannerParam.DATABASE_ID],
                                                             pool=session_pool)
                else:
                    # no pools available (local emulation is being used)
                    self._database = self._instance.database(database_id=conn_data[SpannerParam.DATABASE_ID])

                from google.cloud.spanner_v1 import __version__
                conn_data[DbParam.VERSION] = __version__
            except Exception as e:
                err_msg = exc_format(exc=e,
                                     exc_info=sys.exc_info())
        if err_msg:
            result = False
            if _DB_LOGGERS[DbEngine.SPANNER]:
                _DB_LOGGERS[DbEngine.SPANNER].error(msg=err_msg)
            if isinstance(errors, list):
                errors.append(err_msg)

        return result

    def background_ping_and_prime(self) -> None:
        """
        Keep sessions/transactions ready with background pinging.
        """
        # make sure database pool is the proper type
        # noinspection PyProtectedMember
        if isinstance(self._database._pool, PingingPool):  # noqa: SLF001 - private member accessed
            # noinspection PyTypeChecker
            params: dict[SpannerParam, Any] = _DB_CONN_DATA[DbEngine.SPANNER]
            name: str = f"spanner-ping-{params[SpannerParam.INSTANCE_ID]}/{params[SpannerParam.DATABASE_ID]}"

            # build and launch the pinging daemon
            self._pinger = Thread(target=self._background_ping_and_prime,
                                  name=name,
                                  daemon=True)
            self._pinger.start()

    def _background_ping_and_prime(self) -> None:
        """
        Ping sessions to keep them ready.
        """
        # retrieve the pinging interval
        # noinspection PyTypeChecker
        params: dict[SpannerParam, Any] = _DB_CONN_DATA[DbEngine.SPANNER]
        interval: int = params[SpannerParam.SESSION_PING_INTERVAL]

        # retrieve the database pool
        # noinspection PyProtectedMember
        db_pool: PingingPool = self._database._pool  # noqa: SLF001 - private member accessed

        # background-ping the sessions
        while True:
            with suppress(Exception):
                db_pool.ping()
            time.sleep(interval)


class SpannerConnection:
    """
    An implementation of the RDBMS *connection* paradigm, as applied to the Google Cloud Spanner database.

    These are the instance variables:
        - *_database*: *google.cloud.spanner_v1.database import Database* - the reference database
        - *_db_pool*: session pool associated with the database
        - *_pool_session*: *google.cloud.spanner_v1.session.Session* - session obtained from the pool
        - *_session_txn*: *google.cloud.spanner_v1.transaction.Transaction* - session transaction
        - *_autocommit*: whether the connection is in autocommit mode
    """

    def __init__(self,
                 autocommit: bool = None,
                 errors: list[str] = None) -> None:
        """
        Instantiate a *connection* with a Cloud Spanner database.

        The parameter *autocommit* indicates whether the connection should operate in autocommit mode.
        Regardless of its setting, the connection automatically falls back to autocommit mode if no
        session pools are available.

        :param autocommit: whether the connection is to operate in autocommit mode (defaults to *False*)
        :param errors: incidental error messages (might be a non-empty list)
        """
        # instance variables
        self._database: Database | None = None
        self._db_pool: AbstractSessionPool | None = None
        self._pool_session: Session | None = None
        self._session_txn: Transaction | None = None
        self._autocommit: bool | None = None

        spanner_data: dict[str, Any] = _DB_CONN_DATA.get(DbEngine.SPANNER) or {}
        google_spanner: GoogleSpanner = spanner_data.get(SpannerParam.ENGINE)
        if google_spanner:
            # noinspection PyProtectedMember
            self._database = google_spanner._database  # noqa: SLF001 - private member accessed

            self._autocommit = autocommit
            if not self._autocommit:
                try:
                    # retrieve the database pool
                    # noinspection PyProtectedMember
                    self._db_pool = self._database._pool  # noqa: SLF001 - private member accessed
                    if self._db_pool is None:
                        # autocommit mode is required when pools are not available
                        self._autocommit = True
                    else:
                        # acquire a session from the pool
                        self._pool_session = self._db_pool.get()
                except Exception as e:
                    exc_err: str = exc_format(exc=e,
                                              exc_info=sys.exc_info())
                    if _DB_LOGGERS[DbEngine.SPANNER]:
                        _DB_LOGGERS[DbEngine.SPANNER].error(msg=exc_err)
                    if isinstance(errors, list):
                        errors.append(exc_err)

    def select(self,
               sel_stmt: str,
               param_values: dict[str, Any] = None,
               param_types: dict[str, Type] = None,
               committable: bool = None,
               errors: list[str] = None) -> list[tuple] | None:
        """
        Query the database and return all tuples that satisfy the *sel_stmt* command.

        The command can optionally contain selection criteria, with respective names and values
        given in *param_values*, and names and types given in *param_types*.

        If the connection is in *autocommit* mode, a commit or rollback operation is always performed following
        the query execution. Otherwise, this behaviour is defined by *commitable* (defaults to no commit/rollback).

        :param sel_stmt: SELECT command for the query operation
        :param param_values: the selection criteria, specified as key-value pairs
        :param param_types: the types involved in the selection criteria, specified as key-value pairs
        :param committable: whether to commit or rollback the operation, upon completion
        :param errors: incidental error messages (might be a non-empty list)
        :return: list of tuples containing the search result, *[]* on empty search, or *None* if error
        """
        # initialize the return variable
        result: list[tuple] | None = None

        if self._autocommit:
            try:
                with self._database.snapshot() as snapshot:
                    rows: StreamedResultSet = snapshot.execute_sql(sql=sel_stmt,
                                                                   params=param_values,
                                                                   param_types=param_types)
                    if rows is not None:
                        result = list(rows)
            except Exception as e:
                exc_err: str = exc_format(exc=e,
                                          exc_info=sys.exc_info())
                if isinstance(errors, list):
                    errors.append(exc_err)
        else:
            if self._session_txn is None:
                self._session_txn = Transaction(session=self._pool_session)
                self._session_txn.begin()
            try:
                rows: StreamedResultSet = self._session_txn.execute_sql(sql=sel_stmt,
                                                                        params=param_values,
                                                                        param_types=param_types)
                if rows is not None:
                    result = list(rows)
                if committable:
                    self._session_txn.commit()
                    self._session_txn = None
            except Exception as e:
                exc_err: str = exc_format(exc=e,
                                          exc_info=sys.exc_info())
                if _DB_LOGGERS[DbEngine.SPANNER]:
                    _DB_LOGGERS[DbEngine.SPANNER].error(msg=exc_err)
                if isinstance(errors, list):
                    errors.append(exc_err)
                if committable:
                    self._session_txn.rollback()
                    self._session_txn = None

        return result

    def execute(self,
                exc_stmt: str,
                param_values: dict[str, Any] = None,
                param_types: dict[str, Type] = None,
                committable: bool = None,
                errors: list[str] = None) -> tuple | int | None:
        """
        Execute the command *exc_stmt* on the database.

        This command might be a DML ccommand modifying the database, such as inserting, updating or
        deleting tuples, or it might be a DDL statement, or it might even be an environment-related command.

        If the connection is in *autocommit* mode, a commit or rollback operation is always performed following
        the query execution. Otherwise, this behaviour is defined by *commitable* (defaults to no commit/rollback).

        :param exc_stmt: the command to execute
        :param param_values: the selection criteria, specified as key-value pairs
        :param param_types: the types involved in the selection criteria, specified as key-value pairs
        :param committable: whether to commit or rollback the operation, upon completion
        :param errors: incidental error messages (must be *[]* or *None*)
        :return: the value or values returned by the operation, or *None* if error
        """
        def execute_in_transaction(transaction: Transaction,
                                   curr_stmt: str,
                                   curr_params: dict[str, Any],
                                   curr_types: dict[str, Type]) -> tuple | int | None:

            # initialize the return variable
            result: tuple | int | None = None

            rs: StreamedResultSet = transaction.execute_sql(sql=curr_stmt,
                                                            params=curr_params,
                                                            param_types=curr_types)
            if rs is not None:
                rows: list = list(rs)
                if len(rows) > 0:
                    result = rows[0]

            return result

        # initialize the return variable
        result: tuple | int | None = None

        if self._autocommit:
            try:
                # noinspection PyTypeChecker
                result = self._database.run_in_transaction(func=execute_in_transaction,
                                                           curr_stmt=exc_stmt,
                                                           curr_params=param_values,
                                                           curr_types=param_types)
            except Exception as e:
                exc_err: str = exc_format(exc=e,
                                          exc_info=sys.exc_info())
                if _DB_LOGGERS[DbEngine.SPANNER]:
                    _DB_LOGGERS[DbEngine.SPANNER].error(msg=exc_err)
                if isinstance(errors, list):
                    errors.append(exc_err)
        else:
            if self._session_txn is None:
                self._session_txn = Transaction(session=self._pool_session)
                self._session_txn.begin()
            try:
                rs: StreamedResultSet = self._session_txn.execute_sql(sql=exc_stmt,
                                                                      params=param_values,
                                                                      param_types=param_types)
                if rs is not None:
                    rows: list = list(rs)
                    if len(rows) > 0:
                        result = rows[0]
                if committable:
                    self._session_txn.commit()
                    self._session_txn = None
            except Exception as e:
                exc_err: str = exc_format(exc=e,
                                          exc_info=sys.exc_info())
                if _DB_LOGGERS[DbEngine.SPANNER]:
                    _DB_LOGGERS[DbEngine.SPANNER].error(msg=exc_err)
                if isinstance(errors, list):
                    errors.append(exc_err)
                if committable:
                    self._session_txn.rollback()
                    self._session_txn = None

        return result

    def commit(self) -> None:
        """
        Commit the ongoing transaction.
        """
        if self._session_txn:
            self._session_txn.commit()

    def rollback(self) -> None:
        """
        Rollback the ongoing transaction.
        """
        if self._session_txn:
            self._session_txn.rollback()

    def close(self) -> None:
        """
        Close the connection, releasing its resources.
        """
        # return session to the pool
        if self._pool_session and self._db_pool:
            self._db_pool.put(session=self._pool_session)

        # release all resources
        self._database = None
        self._db_pool = None
        self._pool_session = None
        self._session_txn = None

    def __del__(self) -> None:
        """
        Clean up object before garbage collection.

        Release all resources if 'close()' is not invoked prior to the loss of all references to the object.
        """
        if self._database:
            self.close()



import psycopg2
from contextlib import suppress
from datetime import date, datetime
from logging import Logger
from pathlib import Path
from psycopg2 import Binary
from psycopg2.extras import execute_values
# noinspection PyProtectedMember
# ruff: noqa: N812 - checks for lowercase imports that are aliased to non-lowercase names
from psycopg2._psycopg import connection as Connection
from pypomes_core import (
    DateFormat, DatetimeFormat,
    list_get_coupled, str_as_list, str_between, str_splice
)
from typing import Any, BinaryIO, Final

from .db_common import (
    _DB_LOGGERS, DbEngine, DbParam,
    _assert_query_quota, _build_query_msg, _get_params, _except_msg
)

RESERVED_WORDS: Final[list[str]] = [
    "ALL", "ANALYSE", "ANALYZE", "AND", "ANY", "ARRAY", "AS", "ASC", "ASYMMETRIC", "AUTHORIZATION",
    "BINARY", "BOTH", "CASE", "CAST", "CHECK", "COLLATE", "COLUMN", "CONSTRAINT", "CREATE",
    "CROSS", "CURRENT_CATALOG", "CURRENT_DATE", "CURRENT_ROLE", "CURRENT_SCHEMA", "CURRENT_TIME",
    "CURRENT_TIMESTAMP", "CURRENT_USER", "DEFAULT", "DEFERRABLE", "DESC", "DISTINCT",
    "DO", "ELSE", "END", "EXCEPT", "FALSE", "FETCH", "FOR", "FOREIGN", "FREEZE", "FROM", "FULL",
    "GRANT", "GROUP", "HAVING", "ILIKE", "IN", "INITIALLY", "INNER", "INTERSECT", "INTO", "IS", "ISNULL",
    "JOIN", "LATERAL", "LEADING", "LEFT", "LIKE", "LIMIT", "LOCALTIME", "LOCALTIMESTAMP",
    "NATURAL", "NOT", "NOTNULL", "NULL", "OFFSET", "ON", "ONLY", "OR", "ORDER", "OUTER", "OVER", "OVERLAPS",
    "PLACING", "PRIMARY", "REFERENCES", "RETURNING", "RIGHT", "SELECT", "SESSION_USER", "SIMILAR",
    "SOME", "SYMMETRIC", "TABLE", "THEN", "TO", "TRAILING", "TRUE", "UNION", "UNIQUE", "USER", "USING",
    "VARIADIC", "VERBOSE", "WHEN", "WHERE", "WINDOW", "WITH"
]


def get_connection_string() -> str:
    """
    Build and return the connection string for connecting to the database.

    :return: the connection string
    """
    # retrieve the connection parameters
    db_params: dict[DbParam, Any] = _get_params(DbEngine.POSTGRES)

    # build and return the connection string
    return (f"postgresql+psycopg2://{db_params.get(DbParam.USER)}:"
            f"{db_params.get(DbParam.PWD)}@{db_params.get(DbParam.HOST)}:"
            f"{db_params.get(DbParam.PORT)}/{db_params.get(DbParam.NAME)}")


def connect(autocommit: bool = None,
            errors: list[str] = None) -> Connection | None:
    """
    Obtain and return a connection to the database.

    Return *None* if the connection could not be obtained.

    :param autocommit: whether the connection is to be in autocommit mode (defaults to *False*)
    :param errors: incidental error messages (must be *[]* or *None*)
    :return: the connection to the database, or *None* if error
    """
    # initialize the return variable
    result: Connection | None = None

    # retrieve the connection parameters
    db_params: dict[DbParam, Any] = _get_params(DbEngine.POSTGRES)

    # obtain a connection to the database
    try:
        result = psycopg2.connect(host=db_params.get(DbParam.HOST),
                                  port=db_params.get(DbParam.PORT),
                                  database=db_params.get(DbParam.NAME),
                                  user=db_params.get(DbParam.USER),
                                  password=db_params.get(DbParam.PWD))
        # establish the connection's autocommit mode
        result.autocommit = isinstance(autocommit, bool) and autocommit
    except Exception as e:
        msg: str = _except_msg(exception=e,
                               connection=None,
                               engine=DbEngine.POSTGRES)
        if _DB_LOGGERS[DbEngine.POSTGRES]:
            _DB_LOGGERS[DbEngine.POSTGRES].error(msg=msg)
        if isinstance(errors, list):
            errors.append(msg)

    # log errors
    if _DB_LOGGERS[DbEngine.POSTGRES] and errors:
        _DB_LOGGERS[DbEngine.POSTGRES].error(msg="Error connecting to "
                                                 f"'{db_params.get(DbParam.NAME)}' at '{db_params.get(DbParam.HOST)}'")
    return result


def bind_arguments(stmt: str,
                   bind_vals: list[Any]) -> str:
    """
    Replace the placeholders in *query_stmt* with the values in *bind_vals*, and return the modified query statement.

    Note that using a statement in a situation where values for types other than *bool*, *str*, *int*, *float*,
    *date*, or *datetime* were replaced, may bring about undesirable consequences, as the standard string
    representations for these other types would be used.

    :param stmt: the query statement
    :param bind_vals: the values to replace the placeholders with
    :return: the query statement with the placeholders replaced with their corresponding values
    """
    # initialize the return variable
    result: str = stmt

    # bind the arguments
    for bind_val in bind_vals:
        val: str
        if isinstance(bind_val, bool):
            val = "TRUE" if bind_val else "FALSE"
        elif isinstance(bind_val, int | float):
            val = f"{bind_val}"
        elif isinstance(bind_val, date):
            val = f"TO_TIMESTAMP('{bind_val.strftime(format=DateFormat.INV)}', 'YYYY-MM-DD')"
        elif isinstance(bind_val, datetime):
            val = f"TO_TIMESTAMP('{bind_val.strftime(format=DatetimeFormat.INV)}', 'YYYY-MM-DD H24:MI:SS')"
        else:
            val = f"'{bind_val}'"
        result = result.replace("%s", val, 1)

    return result


def select(sel_stmt: str,
           where_vals: tuple | None,
           min_count: int | None,
           max_count: int | None,
           offset_count: int | None,
           limit_count: int | None,
           connection: Connection | None,
           committable: bool | None,
           errors: list[str] = None) -> list[tuple] | None:
    """
    Query the database and return all tuples that satisfy the *sel_stmt* command.

    The command can optionally contain selection criteria, with respective values given in *where_vals*.
    Care should be exercised if *where_clause* contains *IN* directives. In PostgreSQL, the list of values
    for an attribute with the *IN* directive must be contained in a specific tuple, and the operation will
    break for a list of values containing only 1 element.

    If not positive integers, *min_count*, *max_count*, *offset_count*, and *limit_count* are ignored.
    If both *min_count* and *max_count* are specified with equal values, then exactly that number of
    tuples must be returned by the query. The parameter *offset_count* is used to offset the retrieval
    of tuples, and *limit_count* establishes a ceiling on the number of tuples returned.
    If the search is empty, an empty list is returned.

    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param sel_stmt: SELECT command for the search
    :param where_vals: the values to be associated with the selection criteria
    :param min_count: optionally defines the minimum number of tuples expected
    :param max_count: optionally defines the maximum number of expected
    :param offset_count: number of tuples to skip
    :param limit_count: limit to the number of tuples returned, to be specified in the query statement itself
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (must be *[]* or *None*)
    :return: list of tuples containing the search result, *[]* on empty search, or *None* if error
    """
    # initialize the return variable
    result: list[tuple] | None = None

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # make sure to have a connection
    curr_conn: Connection = connection or connect(autocommit=False,
                                                  errors=errors)
    if not errors:
        # establish offset and limit
        if offset_count or limit_count:
            from . import db_pomes
            sel_stmt = db_pomes.db_add_query_limits(sel_stmt=sel_stmt,
                                                    offset_count=offset_count,
                                                    limit_count=limit_count)
        try:
            # obtain a cursor and execute the operation
            with curr_conn.cursor() as cursor:
                cursor.execute(query=f"{sel_stmt};",
                               vars=where_vals)
                # obtain the number of tuples returned
                rows: list[tuple] = list(cursor)
                count: int = len(rows)

                # log the retrieval operation
                if _DB_LOGGERS[DbEngine.POSTGRES]:
                    from_table: str = str_splice(sel_stmt + " ",
                                                 seps=(" FROM ", " "))[1]
                    _DB_LOGGERS[DbEngine.POSTGRES].debug(msg=f"Read {count} tuples from "
                                                             f"{DbEngine.POSTGRES}.{from_table}, offset "
                                                             f"{offset_count}, connection {id(curr_conn)}")

                if _assert_query_quota(query=sel_stmt,
                                       engine=DbEngine.POSTGRES,
                                       where_vals=where_vals,
                                       count=count,
                                       min_count=min_count,
                                       max_count=max_count,
                                       errors=errors):
                    # the query quota was satisfied, retrieve the returned tuples
                    result = rows

            # commit the transaction
            if committable or not connection:
                curr_conn.commit()
        except Exception as e:
            if committable or not connection:
                with suppress(Exception):
                    curr_conn.rollback()
            msg: str = _except_msg(exception=e,
                                   connection=curr_conn,
                                   engine=DbEngine.POSTGRES)
            if _DB_LOGGERS[DbEngine.POSTGRES]:
                _DB_LOGGERS[DbEngine.POSTGRES].error(msg=msg)
            if isinstance(errors, list):
                errors.append(msg)
        finally:
            # close the connection, if locally acquired
            if curr_conn and not connection:
                curr_conn.close()
        # log errors
        if _DB_LOGGERS[DbEngine.POSTGRES] and errors:
            _DB_LOGGERS[DbEngine.POSTGRES].error(msg=_build_query_msg(query_stmt=sel_stmt,
                                                                      engine=DbEngine.POSTGRES,
                                                                      bind_vals=where_vals))
    return result


def execute(exc_stmt: str,
            bind_vals: tuple | None,
            return_cols: dict[str, type] | None,
            min_count: int | None,
            max_count: int | None,
            connection: Connection | None,
            committable: bool | None,
            errors: list[str] = None) -> tuple | int | None:
    """
    Execute the command *exc_stmt* on the database.

    This command might be a DML ccommand modifying the database, such as inserting, updating or
    deleting tuples, or it might be a DDL statement, or it might even be an environment-related command.

    The optional bind values for this operation are in *bind_vals*. The optional *return_cols* indicate that
    the values of the columns therein should be returned upon execution of *exc_stmt*. This is typical for
    *INSERT* or *UPDATE* statements on tables with *identity-type* columns, which are columns whose values
    are generated by the database itself. Otherwise, the value returned is the number of inserted, modified,
    or deleted tuples, or *None* if an error occurred.

    If not positive integers, *min_count* and *max_count* are ignored. If both *min_count* and *max_count*
    are specified with equal values, then exactly that number of tuples must have been affected by the query,
    as reported by *cursor.rowcount*.

    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param exc_stmt: the command to execute
    :param bind_vals: optional bind values
    :param return_cols: optional columns and respective types, whose values are to be returned on *INSERT* or *UPDATE*
    :param min_count: optionally defines the minimum number of tuples to be affected
    :param max_count: optionally defines the maximum number of tuples to be affected
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (must be *[]* or *None*)
    :return: the values of *return_cols*, the value returned by the operation, or *None* if error
    """
    # initialize the return variable
    result: tuple | int | None = None

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # make sure to have a connection
    curr_conn: Connection = connection or connect(autocommit=False,
                                                  errors=errors)
    if not errors:
        if return_cols:
            exc_stmt += F" RETURNING {', '.join(return_cols.keys())}"
        try:
            # obtain a cursor and execute the operation
            with curr_conn.cursor() as cursor:
                cursor.execute(query=f"{exc_stmt};",
                               vars=bind_vals)

                # has the query quota been satisfied ?
                count: int = cursor.rowcount
                if _assert_query_quota(query=exc_stmt,
                                       engine=DbEngine.POSTGRES,
                                       where_vals=None,
                                       count=count,
                                       min_count=min_count,
                                       max_count=max_count,
                                       errors=errors):
                    if return_cols:
                        result = tuple(cursor.fetchone())
                    else:
                        result = cursor.rowcount

            # commit the transaction, if appropriate
            if committable or not connection:
                curr_conn.commit()
        except Exception as e:
            if committable or not connection:
                with suppress(Exception):
                    curr_conn.rollback()
            msg: str = _except_msg(exception=e,
                                   connection=curr_conn,
                                   engine=DbEngine.POSTGRES)
            if _DB_LOGGERS[DbEngine.POSTGRES]:
                _DB_LOGGERS[DbEngine.POSTGRES].error(msg=msg)
            if isinstance(errors, list):
                errors.append(msg)
        finally:
            # close the connection, if locally acquired
            if curr_conn and not connection:
                curr_conn.close()
        # log errors
        if _DB_LOGGERS[DbEngine.POSTGRES] and errors:
            _DB_LOGGERS[DbEngine.POSTGRES].error(msg=_build_query_msg(query_stmt=exc_stmt,
                                                                      engine=DbEngine.POSTGRES,
                                                                      bind_vals=bind_vals))
    return result


def bulk_execute(exc_stmt: str,
                 exc_vals: list[tuple],
                 template: str | None,
                 connection: Connection | None,
                 committable: bool | None,
                 errors: list[str] = None) -> int | None:
    """
    Bulk-update the database with the statement defined in *execute_stmt*, and the values in *execute_vals*.

    *DELETE* operations require a special syntax using a *IN* clause:
        DELETE FROM my_schema.my_table WHERE (id1, id2) IN (%s)

    For *INSERT* operations, the *VALUES* clause must be simply *VALUES %s*:
        INSERT INTO my_schema.my_table (v1, v2, ...) VALUES %s

    *UPDATE* operations require a special syntax, with *VALUES %s* combined with a *FROM* clause:
        UPDATE my_schema.my_table SET v1 = data.v1, v2 = data.v2, ...
        FROM (VALUES %s) AS data (id, v1, , ...) WHERE my_schema.my_table.id = data.id

    Those special query syntaxes for *INSERT* and *UPDATE* operations present a distinct problem.
    If most or all values passed for a given column are null values, Postgres will implicitly consider them
    to be *TEXT*, and if the column's type is incompatible, an error message is returned due to the resulting
    data type mismatch. The solution is two-fold:
      - for *INSERTs*: add a *template* parameter, casting the values of the columns
      - for *UPDATEs*: explicitly cast the value of the column in the corresponding query clauses

    For illustration purposes, here are an example of the *UPDATE* query statement:
            UPDATE my_schema.my_table SET v1 = data.v1::numeric, v2 = data.v2::timestamp, ...
            FROM (VALUES %s) AS data (id, v1, v2, ...) WHERE my_schema.my_table.id = data.id

    The data types (*int4*, *numeric*, *timestamp*, etc.) may be obtained by querying the columns' metadata.
    Enriching the query statements in this manner is conveniently made available in *add_types()* below.

    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param exc_stmt: the command to update the database with
    :param exc_vals: the list of values for tuple identification, and to update the database with
    :param template: the snippet to merge to every item in *exc_vals* to compose the query
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (must be *[]* or *None*)
    :return: the number of inserted or updated tuples, or *None* if error
    """
    # initialize the return variable
    result: int | None = None

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # make sure to have a connection
    curr_conn: Connection = connection or connect(autocommit=False,
                                                  errors=errors)
    if not errors:
        # execute the bulk query
        try:
            # obtain a cursor and perform the operation
            with curr_conn.cursor() as cursor:
                # 'cursor.rowcount' might end up with a wrong value
                execute_values(cur=cursor,
                               sql=exc_stmt,
                               argslist=exc_vals,
                               template=template)
                result = len(exc_vals)

            # commit the transaction, if appropriate
            if committable or not connection:
                curr_conn.commit()
        except Exception as e:
            if committable or not connection:
                with suppress(Exception):
                    curr_conn.rollback()
            msg: str = _except_msg(exception=e,
                                   connection=curr_conn,
                                   engine=DbEngine.POSTGRES)
            if _DB_LOGGERS[DbEngine.POSTGRES]:
                _DB_LOGGERS[DbEngine.POSTGRES].error(msg=msg)
            if isinstance(errors, list):
                errors.append(msg)
        finally:
            # close the connection, if locally acquired
            if curr_conn and not connection:
                curr_conn.close()
        # log errors
        if _DB_LOGGERS[DbEngine.POSTGRES] and errors:
            _DB_LOGGERS[DbEngine.POSTGRES].error(msg=_build_query_msg(query_stmt=exc_stmt,
                                                                      engine=DbEngine.POSTGRES,
                                                                      bind_vals=exc_vals[0]))
    return result


def update_lob(lob_table: str,
               lob_column: str,
               pk_columns: list[str],
               pk_vals: tuple,
               lob_data: bytes | str | Path | BinaryIO,
               chunk_size: int,
               connection: Connection | None,
               committable: bool | None,
               errors: list[str] = None,
               logger: Logger = None) -> None:
    """
    Update a large binary object (LOB) in the given table and column.

    The data for the update may come from *bytes*, from a *Path* or its string representation, or from
    a pointer obtained from *BytesIO* or *Path.open()* in binary mode.

    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param lob_table: the table to be update with the new LOB
    :param lob_column: the column to be updated with the new LOB
    :param pk_columns: columns making up a primary key, or a unique identifier for the tuple
    :param pk_vals: values with which to locate the tuple to be updated
    :param lob_data: the LOB data (bytes, a file path, or a file pointer)
    :param chunk_size: size in bytes of the data chunk to read/write, or 0 or *None* for no limit
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (must be *[]* or *None*)
    :param logger: optional logger
    """
    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # make sure to have a connection
    curr_conn: Connection = connection or connect(autocommit=False,
                                                  errors=errors)
    if not errors:
        if isinstance(lob_data, str):
            lob_data = Path(lob_data)

        # normalize the chunk size
        if not chunk_size:
            chunk_size = -1

        # build the UPDATE query
        where_clause: str = " AND ".join([f"{column} = %s" for column in pk_columns])
        update_stmt: str = (f"UPDATE {lob_table} "
                            f"SET {lob_column} = %s "
                            f"WHERE {where_clause}")
        try:
            # obtain a cursor and execute the operation
            with curr_conn.cursor() as cursor:

                # retrieve the lob data and write to the database
                if isinstance(lob_data, bytes):
                    cursor.execute(update_stmt,
                                   (Binary(lob_data), *pk_vals))
                elif isinstance(lob_data, Path):
                    data_bytes: bytes
                    with lob_data.open("rb") as file:
                        data_bytes = file.read(chunk_size)
                        while data_bytes:
                            cursor.execute(update_stmt,
                                           (Binary(data_bytes), *pk_vals))
                            data_bytes = file.read(chunk_size)
                else:
                    data_bytes: bytes = lob_data.read(chunk_size)
                    while data_bytes:
                        cursor.execute(update_stmt,
                                       (Binary(data_bytes), *pk_vals))
                        data_bytes = lob_data.read(chunk_size)
                    lob_data.close()

            # commit the transaction, if appropriate
            if committable or not connection:
                curr_conn.commit()
        except Exception as e:
            if committable or not connection:
                with suppress(Exception):
                    curr_conn.rollback()
            msg: str = _except_msg(exception=e,
                                   connection=curr_conn,
                                   engine=DbEngine.POSTGRES)
            if logger:
                logger.error(msg=msg)
            if isinstance(errors, list):
                errors.append(msg)
        finally:
            # close the connection, if locally acquired
            if curr_conn and not connection:
                curr_conn.close()
        # log errors
        if logger and errors:
            logger.error(msg=_build_query_msg(query_stmt=update_stmt,
                                              engine=DbEngine.POSTGRES,
                                              bind_vals=pk_vals))


def call_procedure(proc_name: str,
                   proc_vals: tuple | None,
                   connection: Connection | None,
                   committable: bool | None,
                   errors: list[str] = None) -> list[tuple] | None:
    """
    Execute the stored procedure *proc_name*, with the arguments given in *proc_vals*.

    The parameter *committable* defines whether a commit or rollback is performed on the provided *connection*.

    :param proc_name: the name of the stored procedure
    :param proc_vals: the arguments to be passed
    :param connection: optional connection to use (obtains a new one, if not provided)
    :param committable: whether to commit or rollback the operation, upon completion
    :param errors: incidental error messages (must be *[]* or *None*)
    :return: the data returned by the procedure, or *None* if error
    """
    # initialize the return variable
    result: list[tuple] | None = None

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # make sure to have a connection
    curr_conn: Connection = connection or connect(autocommit=False,
                                                  errors=errors)
    if not errors:
        # build the command
        proc_stmt: str = f"{proc_name}(" + "%s, " * (len(proc_vals) - 1) + "%s)"

        # execute the stored procedure
        try:
            # obtain a cursor and perform the operation
            with curr_conn.cursor() as cursor:
                cursor.execute(query=proc_stmt,
                               vars=proc_vals)
                # retrieve the returned tuples
                result = list(cursor)

            # commit the transaction
            if committable or not connection:
                curr_conn.commit()
        except Exception as e:
            if committable or not connection:
                with suppress(Exception):
                    curr_conn.rollback()
            msg: str = _except_msg(exception=e,
                                   connection=curr_conn,
                                   engine=DbEngine.POSTGRES)
            if _DB_LOGGERS[DbEngine.POSTGRES]:
                _DB_LOGGERS[DbEngine.POSTGRES].error(msg=msg)
            if isinstance(errors, list):
                errors.append(msg)
        finally:
            # close the connection, if locally acquired
            if curr_conn and not connection:
                curr_conn.close()
        # log errors
        if _DB_LOGGERS[DbEngine.POSTGRES] and errors:
            _DB_LOGGERS[DbEngine.POSTGRES].error(msg=_build_query_msg(query_stmt=proc_stmt,
                                                                      engine=DbEngine.POSTGRES,
                                                                      bind_vals=proc_vals))
    return result


def identity_post_insert(insert_stmt: str,
                         connection: Connection,
                         committable: bool,
                         identity_column: str,
                         errors: list[str] = None) -> None:
    """
    Handle the post-insert for tables with identity columns.

    Identity columns are columns whose values are generated directly by the database, and as such,
    require special handling before and after bulk inserts.

    :param insert_stmt: the INSERT command
    :param connection: the connection to use
    :param committable: whether to commit or rollback the operation, upon completion
    :param identity_column: column whose values are generated by the database
    :param errors: incidental error messages (must be *[]* or *None*)
    """
    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # obtain the maximum value inserted
    table_name: str = str_between(insert_stmt.upper(),
                                  from_str=" INTO ",
                                  to_str=" ")
    recs: list[tuple[int]] = select(sel_stmt=(f"SELECT MAX({identity_column}) "
                                              f"FROM {table_name}"),
                                    where_vals=None,
                                    min_count=None,
                                    max_count=None,
                                    offset_count=None,
                                    limit_count=None,
                                    connection=connection,
                                    committable=False,
                                    errors=errors)
    if not errors:
        sel_stmt: str = (f"SELECT setval("
                         f"pg_get_serial_sequence('{table_name}', '{identity_column}'), "
                         f"{recs[0][0]})")
        execute(exc_stmt=sel_stmt,
                bind_vals=None,
                return_cols=None,
                min_count=None,
                max_count=None,
                connection=connection,
                committable=committable,
                errors=errors)


def build_typified_template(insert_stmt: str,
                            nullable_only: bool,
                            connection: Connection,
                            errors: list[str] = None) -> str:
    """
    Build the typified template corresponding to the columns in *insert_stmt*, by setting the appropriate data types.

    As an illustration, the statement
        INSERT INTO my_schema.my_table (v1, v2, ...) VALUES %s

    would yield the template
        (%s:int4, %s:timestamp, ...)

    depending on the nullability and types of the columns, and on the value of *nullable_only*.

    :param insert_stmt: the bulk *INSERT* statement
    :param nullable_only: whether to disregard non-nullable columns
    :param connection: the connection to use
    :param errors: incidental error messages (must be *[]* or *None*)
    :return: the typified template, or *None* if error or no column was typified
    """
    # initialize the return variable
    result: str | None = None

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # retrieve the table name and schema
    table_name: str = insert_stmt[12:insert_stmt.find(" (")]
    table_schema: str
    table_schema, table_name = table_name.split(sep=".") if "." in table_name else (None, table_name)

    # obtain the columns' metadata
    sel_stmt: str = ("SELECT column_name, udt_name "
                     "FROM information_schema.columns "
                     f"WHERE table_name = '{table_name}'")
    if table_schema:
        sel_stmt += f" AND table_schema = '{table_schema}'"
    if nullable_only:
        sel_stmt += " AND is_nullable = 'YES'"
    # noinspection PyTypeChecker
    recs: list[tuple[str, str]] = select(sel_stmt=sel_stmt,
                                         where_vals=None,
                                         min_count=None,
                                         max_count=None,
                                         offset_count=None,
                                         limit_count=None,
                                         connection=connection,
                                         committable=False,
                                         errors=errors)
    # build the template
    if not errors:
        result = "("
        # obtain the columns in the insert statement
        columns_clause: str = insert_stmt[insert_stmt.index("(")+1:insert_stmt.index(")")]
        columns: list[str] = str_as_list(columns_clause)
        for column in columns:
            result += "%s"
            data_type: str = list_get_coupled(coupled_elements=recs,
                                              primary_element=column)
            if data_type:
                result += f"::{data_type}"
            result += ", "
        if "::" in result:
            result = result[:-2] + ")"
        else:
            result = None

    return result


def tipify_bulk_update(update_stmt: str,
                       nullable_only: bool,
                       connection: Connection,
                       errors: list[str] = None) -> str:
    """
    Modify the bulk *update_stmt* statement by adding the appropriate data types.

    As an illustration, the statement
      - UPDATE my_schema.my_table SET v1 = data.v1, v2 = data.v2, ...
        FROM (VALUES %s) AS data (id, v1, v2) WHERE my_schema.my_table.id = data.id

    would result in the statement
      - UPDATE my_schema.my_table SET v1 = data.v1::int4, v2 = data.v2::timestamp, ...
        FROM (VALUES %s) AS data (id, v1, v2) WHERE my_schema.my_table.id = data.id

    depending on the nullability and types of the columns, and on the value of *nullable_only*.

    :param update_stmt: the bulk *UPDATE* statement
    :param nullable_only: whether to disregard non-nullable columns
    :param connection: the connection to use
    :param errors: incidental error messages (must be *[]* or *None*)
    :return: a new statement, enriched with the columns' data types, or *None* if error
    """
    # initialize the return variable
    result: str | None = None

    # make sure to have an errors list
    if not isinstance(errors, list):
        errors = []

    # retrieve the table name and schema
    table_name: str = update_stmt[7:update_stmt.find(" SET")]
    table_schema: str
    table_schema, table_name = table_name.split(sep=".") if "." in table_name else (None, table_name)

    # obtain the columns' metadata
    sel_stmt: str = ("SELECT column_name, udt_name "
                     "FROM information_schema.columns "
                     f"WHERE table_name = '{table_name}'")
    if table_schema:
        sel_stmt += f" AND table_schema = '{table_schema}'"
    if nullable_only:
        sel_stmt += " AND is_nullable = 'YES'"
    # noinspection PyTypeChecker
    recs: list[tuple[str, str]] = select(sel_stmt=sel_stmt,
                                         where_vals=None,
                                         min_count=None,
                                         max_count=None,
                                         offset_count=None,
                                         limit_count=None,
                                         connection=connection,
                                         committable=False,
                                         errors=errors)
    # tipify the columns
    if not errors:
        result = update_stmt
        for rec in recs:
            result = result.replace(f".{rec[0]},", f".{rec[0]}::{rec[1]},", 1)
            result = result.replace(f".{rec[0]} FROM", f".{rec[0]}::{rec[1]} FROM", 1)

    return result
